@@ -3,7 +3,6 @@
 #endif
 
 #include <math.h>
-#include <libgnome/gnome-macros.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <gdk/gdkkeysyms.h>
 
@@ -82,6 +81,9 @@ struct _EogScrollViewPrivate {
 	/* the actual zoom factor */
 	double zoom;
 
+	/* the minimum possible (reasonable) zoom factor */
+	double min_zoom;
+
 	/* Current scrolling offsets */
 	int xofs, yofs;
 
@@ -115,11 +117,9 @@ struct _EogScrollViewPrivate {
 static void scroll_by (EogScrollView *view, int xofs, int yofs);
 static void set_zoom_fit (EogScrollView *view);
 static void request_paint_area (EogScrollView *view, GdkRectangle *area);
+static void set_minimum_zoom_factor (EogScrollView *view);
 
-GNOME_CLASS_BOILERPLATE (EogScrollView,
-			 eog_scroll_view,
-			 GtkTable,
-			 GTK_TYPE_TABLE);
+G_DEFINE_TYPE (EogScrollView, eog_scroll_view, GTK_TYPE_TABLE)
 
 
 /*===================================
@@ -361,7 +361,8 @@ check_scrollbar_visibility (EogScrollView *view, GtkAllocation *alloc)
 	}
 }
 
-#define DOUBLE_EQUAL(a,b) (fabs (a - b) < 1e-6)
+#define DOUBLE_EQUAL_MAX_DIFF 1e-6
+#define DOUBLE_EQUAL(a,b) (fabs (a - b) < DOUBLE_EQUAL_MAX_DIFF)
 
 /* Returns whether the zoom factor is 1.0 */
 static gboolean
@@ -388,6 +389,7 @@ is_image_movable (EogScrollView *view)
 
 
 /* Computes the image offsets with respect to the window */
+/*
 static void
 get_image_offsets (EogScrollView *view, int *xofs, int *yofs)
 {
@@ -402,7 +404,7 @@ get_image_offsets (EogScrollView *view, int *xofs, int *yofs)
 	width = GTK_WIDGET (priv->display)->allocation.width;
 	height = GTK_WIDGET (priv->display)->allocation.height;
 
-	/* Compute image offsets with respect to the window */
+	// Compute image offsets with respect to the window
 	if (scaled_width <= width)
 		*xofs = (width - scaled_width) / 2;
 	else
@@ -413,8 +415,8 @@ get_image_offsets (EogScrollView *view, int *xofs, int *yofs)
 	else
 		*yofs = -priv->yofs;
 }
+*/
 
-
 /*===================================
           drawing core
   ---------------------------------*/
@@ -484,7 +486,7 @@ paint_rectangle (EogScrollView *view, ArtIRect *rect, GdkInterpType interp_type)
 		paint_background (view, &r, rect);
 		return;
 	}
-
+	
 	/* Compute image offsets with respect to the window */
 
 	if (scaled_width <= width)
@@ -905,12 +907,21 @@ drag_to (EogScrollView *view, int x, int y)
 	scroll_to (view, x, y, TRUE);
 }
 
+static void
+set_minimum_zoom_factor (EogScrollView *view)
+{
+	g_return_if_fail (EOG_IS_SCROLL_VIEW (view));
+	
+	view->priv->min_zoom = MAX (1.0 / gdk_pixbuf_get_width (view->priv->pixbuf),
+				    MAX(1.0 / gdk_pixbuf_get_height (view->priv->pixbuf),
+					MIN_ZOOM_FACTOR) );
+	return;
+}
 
 /**
  * set_zoom:
  * @view: A scroll view.
- * @zoomx: Horizontal zoom factor.
- * @zoomy: Vertical zoom factor.
+ * @zoom: Zoom factor.
  * @have_anchor: Whether the anchor point specified by (@anchorx, @anchory)
  * should be used.
  * @anchorx: Horizontal anchor point in pixels.
@@ -945,6 +956,8 @@ set_zoom (EogScrollView *view, double zoom,
 
 	if (DOUBLE_EQUAL (priv->zoom, zoom))
 		return;
+	if (DOUBLE_EQUAL (priv->zoom, priv->min_zoom) && zoom < priv->zoom)
+		return;
 
 	priv->zoom_mode = ZOOM_MODE_FREE;
 
@@ -971,7 +984,10 @@ set_zoom (EogScrollView *view, double zoom,
 #if 0
 	g_print ("xofs: %i  yofs: %i\n", priv->xofs, priv->yofs);
 #endif
-	priv->zoom = zoom;
+	if (zoom <= priv->min_zoom)
+		priv->zoom = priv->min_zoom;
+	else
+		priv->zoom = zoom;
 
 	/* we make use of the new values here */
 	check_scrollbar_visibility (view, NULL);
@@ -1014,9 +1030,6 @@ set_zoom_fit (EogScrollView *view)
 	else if (new_zoom < MIN_ZOOM_FACTOR)
 		new_zoom = MIN_ZOOM_FACTOR;
 
-	if (DOUBLE_EQUAL (new_zoom, priv->zoom))
-		return;
-
 	priv->zoom = new_zoom;
 	priv->xofs = 0;
 	priv->yofs = 0;
@@ -1024,7 +1037,6 @@ set_zoom_fit (EogScrollView *view)
 	g_signal_emit (view, view_signals [SIGNAL_ZOOM_CHANGED], 0, priv->zoom);
 }
 
-
 /*===================================
 
    internal signal callbacks
@@ -1037,22 +1049,23 @@ display_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer data)
 {
 	EogScrollView *view;
 	EogScrollViewPrivate *priv;
-	gboolean handled;
 	gboolean do_zoom;
 	double zoom;
 	gboolean do_scroll;
 	int xofs, yofs;
+	int width, height;
 
 	view = EOG_SCROLL_VIEW (data);
 	priv = view->priv;
 
-	handled = FALSE;
-
 	do_zoom = FALSE;
 	do_scroll = FALSE;
 	xofs = yofs = 0;
 	zoom = 1.0;
 
+	width = GTK_WIDGET (priv->display)->allocation.width;
+	height = GTK_WIDGET (priv->display)->allocation.height;
+
 	switch (event->keyval) {
 	case GDK_Up:
 		do_scroll = TRUE;
@@ -1060,12 +1073,34 @@ display_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer data)
 		yofs = -SCROLL_STEP_SIZE;
 		break;
 
+	case GDK_Page_Up:
+		do_scroll = TRUE;
+		if (event->state & GDK_CONTROL_MASK) {
+			xofs = -(width * 3) / 4;
+			yofs = 0;
+		} else {
+			xofs = 0;
+			yofs = -(height * 3) / 4;
+		}
+		break;
+
 	case GDK_Down:
 		do_scroll = TRUE;
 		xofs = 0;
 		yofs = SCROLL_STEP_SIZE;
 		break;
 
+	case GDK_Page_Down:
+		do_scroll = TRUE;
+		if (event->state & GDK_CONTROL_MASK) {
+			xofs = (width * 3) / 4;
+			yofs = 0;
+		} else {
+			xofs = 0;
+			yofs = (height * 3) / 4;
+		}
+		break;
+
 	case GDK_Left:
 		do_scroll = TRUE;
 		xofs = -SCROLL_STEP_SIZE;
@@ -1138,22 +1173,23 @@ eog_scroll_view_button_press_event (GtkWidget *widget, GdkEventButton *event, gp
 		return FALSE;
 
 	switch (event->button) {
-	case 1:
-		if (is_image_movable (view)) {
-			priv->cursor_type = CURSOR_HAND_CLOSED;
-			cursor_set (GTK_WIDGET (priv->display), CURSOR_HAND_CLOSED);
+		case 1:
+		case 2:
+			if (is_image_movable (view)) {
+				priv->cursor_type = CURSOR_HAND_CLOSED;
+				cursor_set (GTK_WIDGET (priv->display), CURSOR_HAND_CLOSED);
 
-			priv->dragging = TRUE;
-			priv->drag_anchor_x = event->x;
-			priv->drag_anchor_y = event->y;
-			
-			priv->drag_ofs_x = priv->xofs;
-			priv->drag_ofs_y = priv->yofs;
+				priv->dragging = TRUE;
+				priv->drag_anchor_x = event->x;
+				priv->drag_anchor_y = event->y;
 
-			return TRUE;
-		}
-	default:
-		break;
+				priv->drag_ofs_x = priv->xofs;
+				priv->drag_ofs_y = priv->yofs;
+
+				return TRUE;
+			}
+		default:
+			break;
 	}
 
 	return FALSE;
@@ -1182,19 +1218,26 @@ eog_scroll_view_button_release_event (GtkWidget *widget, GdkEventButton *event,
 	view = EOG_SCROLL_VIEW (data);
 	priv = view->priv;
 
-	if (!priv->dragging || event->button != 1)
+	if (!priv->dragging)
 		return FALSE;
 
-	drag_to (view, event->x, event->y);
-	priv->dragging = FALSE;
+	switch (event->button) {
+		case 1:
+		case 2:
+			drag_to (view, event->x, event->y);
+			priv->dragging = FALSE;
 
-	if (is_image_movable (view)) {
-		priv->cursor_type = CURSOR_HAND_OPEN;
-		cursor_set (GTK_WIDGET (priv->display), CURSOR_HAND_OPEN);
-	}
-	else {
-		priv->cursor_type = CURSOR_DEFAULT;
-		cursor_set (GTK_WIDGET (priv->display), CURSOR_DEFAULT);
+			if (is_image_movable (view)) {
+				priv->cursor_type = CURSOR_HAND_OPEN;
+				cursor_set (GTK_WIDGET (priv->display), CURSOR_HAND_OPEN);
+			}
+			else {
+				priv->cursor_type = CURSOR_DEFAULT;
+				cursor_set (GTK_WIDGET (priv->display), CURSOR_DEFAULT);
+			}
+
+		default:
+			break;
 	}
 
 	return TRUE;
@@ -1314,7 +1357,8 @@ eog_scroll_view_size_allocate (GtkWidget *widget, GtkAllocation *alloc)
 	view = EOG_SCROLL_VIEW (widget);
 	check_scrollbar_visibility (view, alloc);
 
-	GNOME_CALL_PARENT (GTK_WIDGET_CLASS, size_allocate, (widget, alloc));
+	GTK_WIDGET_CLASS (eog_scroll_view_parent_class)->size_allocate (widget
+									,alloc);
 }
 
 static void
@@ -1409,7 +1453,7 @@ display_expose_event (GtkWidget *widget, GdkEventExpose *event, gpointer data)
    image loading callbacks
 
    -----------------------------------*/
-
+/*
 static void
 image_loading_update_cb (EogImage *img, int x, int y, int width, int height, gpointer data)
 {
@@ -1470,7 +1514,7 @@ image_loading_finished_cb (EogImage *img, gpointer data)
 	else if (priv->interp_type != GDK_INTERP_NEAREST &&
 		 !is_unity_zoom (view))
 	{
-		/* paint antialiased image version */
+		// paint antialiased image version 
 		priv->progressive_state = PROGRESSIVE_POLISHING;
 		gtk_widget_queue_draw (GTK_WIDGET (priv->display));
 	}
@@ -1511,7 +1555,7 @@ image_loading_cancelled_cb (EogImage *img, gpointer data)
 		gdk_window_clear (GTK_WIDGET (priv->display)->window);
 	}
 }
-
+*/
 static void
 image_changed_cb (EogImage *img, gpointer data)
 {
@@ -1525,7 +1569,6 @@ image_changed_cb (EogImage *img, gpointer data)
 	}
 
 	priv->pixbuf = eog_image_get_pixbuf (img);
-	g_object_ref (priv->pixbuf);
 
 	set_zoom_fit (EOG_SCROLL_VIEW (data));
 	check_scrollbar_visibility (EOG_SCROLL_VIEW (data), NULL);
@@ -1638,7 +1681,8 @@ eog_scroll_view_zoom_in (EogScrollView *view, gboolean smooth)
 		int index = -1;
 
 		for (i = 0; i < n_zoom_levels; i++) {
-			if (preferred_zoom_levels [i] > priv->zoom) {
+			if (preferred_zoom_levels [i] - priv->zoom
+					> DOUBLE_EQUAL_MAX_DIFF) {
 				index = i;
 				break;
 			}
@@ -1673,7 +1717,8 @@ eog_scroll_view_zoom_out (EogScrollView *view, gboolean smooth)
 		int index = -1;
 
 		for (i = n_zoom_levels - 1; i >= 0; i--) {
-			if (preferred_zoom_levels [i] < priv->zoom) {
+			if (priv->zoom - preferred_zoom_levels [i]
+					> DOUBLE_EQUAL_MAX_DIFF) {
 				index = i;
 				break;
 			}
@@ -1719,7 +1764,10 @@ eog_scroll_view_get_zoom_is_min (EogScrollView *view)
 {
 	g_return_val_if_fail (EOG_IS_SCROLL_VIEW (view), FALSE);
 
-	return DOUBLE_EQUAL (view->priv->zoom, MIN_ZOOM_FACTOR);
+	set_minimum_zoom_factor (view);
+
+	return DOUBLE_EQUAL (view->priv->zoom, MIN_ZOOM_FACTOR) || 
+	       DOUBLE_EQUAL (view->priv->zoom, view->priv->min_zoom);
 }
 
 gboolean
@@ -1797,6 +1845,15 @@ eog_scroll_view_get_image_size   (EogScrollView *view, int *width, int *height,
 	*height = gdk_pixbuf_get_height (priv->pixbuf);
 }
 
+gboolean 
+eog_scroll_view_scrollbars_visible (EogScrollView *view)
+{
+	if (!GTK_WIDGET_VISIBLE (GTK_WIDGET (view->priv->hbar)) && 
+	    !GTK_WIDGET_VISIBLE (GTK_WIDGET (view->priv->vbar)))
+		return FALSE;
+
+	return TRUE;
+}
 
 /*===================================
     object creation/freeing
@@ -1804,7 +1861,7 @@ eog_scroll_view_get_image_size   (EogScrollView *view, int *width, int *height,
 
 
 static void
-eog_scroll_view_instance_init (EogScrollView *view)
+eog_scroll_view_init (EogScrollView *view)
 {
 	EogScrollViewPrivate *priv;
 
@@ -1812,6 +1869,7 @@ eog_scroll_view_instance_init (EogScrollView *view)
 	view->priv = priv;
 
 	priv->zoom = 1.0;
+	priv->min_zoom = MIN_ZOOM_FACTOR;
 	priv->zoom_mode = ZOOM_MODE_FIT;
 	priv->upscale = FALSE;
 	priv->uta = NULL;
@@ -1847,7 +1905,7 @@ eog_scroll_view_dispose (GObject *object)
 
 	free_image_resources (view);
 
-	GNOME_CALL_PARENT (G_OBJECT_CLASS, dispose, (object));
+	G_OBJECT_CLASS (eog_scroll_view_parent_class)->dispose (object);
 }
 
 static void
@@ -1861,7 +1919,7 @@ eog_scroll_view_finalize (GObject *object)
 		view->priv = 0;
 	}
 
-	GNOME_CALL_PARENT (G_OBJECT_CLASS, finalize, (object));
+	G_OBJECT_CLASS (eog_scroll_view_parent_class)->finalize (object);
 }
 
 static void
@@ -1901,6 +1959,7 @@ eog_scroll_view_new (void)
 	EogScrollViewPrivate *priv;
 
 	widget = g_object_new (EOG_TYPE_SCROLL_VIEW,
+			       "can-focus", TRUE,
 			       "n_rows", 2,
 			       "n_columns", 2,
 			       "homogeneous", FALSE,
@@ -1962,6 +2021,6 @@ eog_scroll_view_new (void)
 			  0, 0);
 
 	gtk_widget_show_all (widget);
-
+	
 	return widget;
 }
